@@ -8,8 +8,8 @@ import math
 import operator
 import re
 import sys
-import unittest
 from abc import ABC, abstractmethod
+from array import array
 from numbers import Integral
 from typing import (
     Callable,
@@ -32,6 +32,7 @@ from typing import (
     Type,
     TypeVar,
     Union,
+    get_type_hints,
 )
 
 import pytest
@@ -39,7 +40,6 @@ import z3  # type: ignore
 
 from crosshair import type_repo
 from crosshair.core import (
-    CrossHairValue,
     analyze_function,
     deep_realize,
     proxy_for_type,
@@ -47,13 +47,16 @@ from crosshair.core import (
     standalone_statespace,
 )
 from crosshair.core_and_libs import run_checkables
+from crosshair.dynamic_typing import origin_of
 from crosshair.libimpl.builtinslib import (
     LazyIntSymbolicStr,
+    ModelingDirector,
+    PreciseIeeeSymbolicFloat,
+    RealBasedSymbolicFloat,
     SymbolicArrayBasedUniformTuple,
     SymbolicBool,
     SymbolicByteArray,
     SymbolicBytes,
-    SymbolicFloat,
     SymbolicInt,
     SymbolicList,
     SymbolicObject,
@@ -68,10 +71,22 @@ from crosshair.statespace import (
     EXEC_ERR,
     POST_FAIL,
     MessageType,
+    StateSpace,
+)
+from crosshair.test_util import (
+    check_exec_err,
+    check_messages,
+    check_states,
+    summarize_execution,
 )
-from crosshair.test_util import check_exec_err, check_states, summarize_execution
 from crosshair.tracers import NoTracing, ResumedTracing
-from crosshair.util import CrosshairInternal, IgnoreAttempt, set_debug
+from crosshair.util import (
+    CrossHairInternal,
+    CrossHairValue,
+    IgnoreAttempt,
+    UnknownSatisfiability,
+    set_debug,
+)
 
 
 class Cat:
@@ -134,14 +149,14 @@ NAN = float("nan")
 
 
 def test_crosshair_types_for_python_type() -> None:
-    assert crosshair_types_for_python_type(int) == (SymbolicInt,)
+    assert crosshair_types_for_python_type(int) == ((SymbolicInt, 1.0),)
     assert crosshair_types_for_python_type(SmokeDetector) == ()
 
 
 def test_isinstance():
     with standalone_statespace:
         with NoTracing():
-            f = SymbolicFloat("f")
+            f = RealBasedSymbolicFloat("f")
         assert isinstance(f, float)
         assert not isinstance(f, int)
 
@@ -149,7 +164,7 @@ def test_isinstance():
 def test_smtfloat_like_a_float():
     with standalone_statespace:
         with NoTracing():
-            f1 = SymbolicFloat("f")
+            f1 = RealBasedSymbolicFloat("f")
         f2 = type(f1)(12)
         with NoTracing():
             assert isinstance(f2, float)
@@ -222,7 +237,7 @@ def test_number_simple_compare_ok() -> None:
     check_states(f, CONFIRMED)
 
 
-def test_number_promotion_compare_unknown() -> None:
+def test_number_promotion_compare_ok() -> None:
     def f(i: int, f: float) -> bool:
         """
         pre: i == 7
@@ -231,7 +246,7 @@ def test_number_promotion_compare_unknown() -> None:
         """
         return i == f and f >= i and i >= f
 
-    check_states(f, CANNOT_CONFIRM)
+    check_states(f, CONFIRMED)
 
 
 def test_numeric_promotions() -> None:
@@ -252,7 +267,7 @@ def test_float_as_bool() -> None:
         """
         return x or y
 
-    check_states(f, CANNOT_CONFIRM)
+    check_states(f, CONFIRMED)
 
 
 def test_int_reverse_operators() -> None:
@@ -329,6 +344,29 @@ def test_int___pow___method():
     check_states(f, POST_FAIL)
 
 
+def test_int___pow___to_ieee_float():
+    with standalone_statespace as space:
+        with NoTracing():
+            space.extra(ModelingDirector).global_representations[
+                float
+            ] = PreciseIeeeSymbolicFloat
+            a = SymbolicInt("a")
+        with pytest.raises(UnknownSatisfiability):
+            sqrt_a = a**0.5
+
+
+def test_int___pow___to_real_based_float():
+    with standalone_statespace as space:
+        with NoTracing():
+            space.extra(ModelingDirector).global_representations[
+                float
+            ] = RealBasedSymbolicFloat
+            a = SymbolicInt("a")
+        sqrt_a = a**0.5
+        with pytest.raises(UnknownSatisfiability):
+            realize(sqrt_a == 3)
+
+
 @pytest.mark.demo
 def test_int___sub___method():
     def f(a: int) -> int:
@@ -418,6 +456,7 @@ def test_int___truediv___method() -> None:
 def test_trunc_fail() -> None:
     def f(n: float) -> int:
         """
+        pre: math.isfinite(n)
         pre: n > 100
         post: _ < n
         """
@@ -453,11 +492,11 @@ def test_round_fail() -> None:
 def test_round_unknown() -> None:
     def f(num: float, ndigits: Optional[int]) -> float:
         """
+        pre: math.isfinite(num)
         post: isinstance(_, int) == (ndigits is None)
         """
         return round(num, ndigits)
 
-    # TODO: this is unknown (rounding reals is hard)
     check_states(f, CANNOT_CONFIRM)
 
 
@@ -466,7 +505,7 @@ def test_float_isinstance() -> None:
         """post: isinstance(_, float)"""
         return x
 
-    check_states(f, CANNOT_CONFIRM)
+    check_states(f, CONFIRMED)
 
 
 def test_mismatched_types() -> None:
@@ -503,15 +542,20 @@ def test_float_from_hex() -> None:
     check_states(f, CONFIRMED)
 
 
-def test_int_from_bytes() -> None:
-    def f(byt: bytes) -> int:
-        """
-        pre: len(byt) == 2
-        post: _ != 5
-        """
-        return int.from_bytes(byt, byteorder="little")
+def test_int_from_byte_iterator(space) -> None:
+    byts = proxy_for_type(bytes, "byts")
+    with ResumedTracing():
+        space.add(len(byts) == 2)
+        number = int.from_bytes(iter(byts), byteorder="little")
+        assert space.is_possible(number == 5)
 
-    check_states(f, POST_FAIL)
+
+def test_int_from_bytes(space) -> None:
+    byts = proxy_for_type(bytes, "byts")
+    with ResumedTracing():
+        space.add(len(byts) == 2)
+        number = int.from_bytes(byts, byteorder="little")
+        assert space.is_possible(number == 5)
 
 
 def test_int_nonlinear() -> None:
@@ -560,19 +604,22 @@ def test_bool_ops(b, op):
     with standalone_statespace as space:
         with NoTracing():
             a = SymbolicBool("a")
-            space.add(a.var)
+        space.add(a)
         symbolic_ret = summarize_execution(lambda: op(a, b))
         concrete_ret = summarize_execution(lambda: op(realize(a), b), detach_path=False)
         assert symbolic_ret == concrete_ret
 
 
 @pytest.mark.parametrize("b", (False, 1, -2.0, NAN, INF, -INF))
-@pytest.mark.parametrize("op", (operator.lt, operator.eq, operator.add, operator.mul))
+@pytest.mark.parametrize(
+    "op",
+    (operator.lt, operator.eq, operator.add, operator.mul, operator.eq, operator.ne),
+)
 def test_float_ops(b, op):
     with standalone_statespace as space:
         with NoTracing():
-            a = SymbolicFloat("a")
-            space.add(a.var < 0)
+            a = space.extra(ModelingDirector).choose(float)("a")
+        space.add(a < 0)
         symbolic_ret = summarize_execution(lambda: op(a, b))
         concrete_ret = summarize_execution(lambda: op(realize(a), b), detach_path=False)
         assert symbolic_ret == concrete_ret
@@ -589,6 +636,26 @@ def test_int_from_str():
     check_states(f, POST_FAIL)
 
 
+def test_int_from_str_with_bases(space):
+    i = proxy_for_type(int, "i")
+    s_11 = proxy_for_type(str, "s_11")
+    s_a = proxy_for_type(str, "s_a")
+    with ResumedTracing():
+        space.add(len(s_11) == 2)
+        space.add(s_11 == "11")
+        space.add(len(s_a) == 1)
+        space.add(s_a == "a")
+        assert int(s_11, 4) == 5
+        assert int(s_11, 16) == 17
+        assert int(s_11, 36) == 37
+        assert int(s_11, 0) == 11
+        assert int(s_a, 16) == 10
+        with pytest.raises(TypeError):
+            assert int(s_a, base=2.5)
+        with pytest.raises(TypeError):
+            int(i, base="foo")
+
+
 def test_easy_float_from_str():
     def f(a: str) -> float:
         """
@@ -612,15 +679,31 @@ def test_float_from_three_digit_str():
                 proxy_for_type(int, "xat1"),
                 proxy_for_type(int, "xat2"),
             ]
-            for point in codepoints:
-                space.add(point.var >= ord("0"))
-                space.add(point.var <= ord("9"))
             x = LazyIntSymbolicStr(codepoints)
+        for point in codepoints:
+            space.add(point >= ord("0"))
+            space.add(point <= ord("9"))
         asfloat = float(x)
-        assert space.is_possible(asfloat.var <= 999)
-        assert not space.is_possible(asfloat.var > 999)
-        assert space.is_possible(asfloat.var == 0)  # (because "000" is a valid float)
-        assert not space.is_possible(asfloat.var == 500.5)
+        assert space.is_possible(asfloat <= 999)
+        assert not space.is_possible(asfloat > 999)
+        assert space.is_possible(asfloat == 0)  # (because "000" is a valid float)
+        assert not space.is_possible(asfloat == 500.5)
+
+
+@pytest.mark.demo("yellow")
+def test_float___pow___operator():
+    def f(a: float) -> float:
+        """
+        Can the given float, cubed, equal 0.125?
+
+        NOTE: Although this example works, nonlinear arithmetic is quite difficult
+        to reason about in most cases.
+
+        post: _ != 0.125
+        """
+        return a**3
+
+    check_states(f, POST_FAIL)
 
 
 def test_int_bitwise_find_negative_input():
@@ -638,7 +721,7 @@ def test_int_bitwise_find_negative_input():
 def test_int_bit_length(val):
     with standalone_statespace as space:
         x = proxy_for_type(int, "x")
-        space.add(x.var == val)
+        space.add(x == val)
         assert realize(x.bit_length()) == val.bit_length()
 
 
@@ -648,7 +731,7 @@ def test_int_bit_length(val):
 def test_int_to_bytes(val):
     with standalone_statespace as space:
         x = proxy_for_type(int, "x")
-        space.add(x.var == val)
+        space.add(x == val)
         assert realize(x.to_bytes(2, "big", signed=True)) == val.to_bytes(
             2, "big", signed=True
         )
@@ -658,7 +741,7 @@ def test_int_format():
     with standalone_statespace as space:
         with NoTracing():
             x = SymbolicInt("x")
-            space.add(x.var == 42)
+        space.add(x == 42)
         assert x.__format__("") == "42"
         # TODO this fails:
         # assert x.__format__("f") == "42.000000"
@@ -1158,18 +1241,17 @@ def test_str_center():
     with standalone_statespace as space:
         with NoTracing():
             string = LazyIntSymbolicStr("string")
-            space.add(string.__len__().var == 3)
             fillch = LazyIntSymbolicStr("fillch")
-            space.add(fillch.__len__().var == 1)
             sz = SymbolicInt("sz")
-            space.add(sz.var > 5)
             sz6 = SymbolicInt("sz6")
-            space.add(sz6.var == 6)
+        space.add(string.__len__() == 3)
+        space.add(fillch.__len__() == 1)
+        space.add(sz > 5)
+        space.add(sz6 == 6)
         assert "boo".center(sz6) == " boo  "
         symbolic_centered = "boo".center(sz, fillch)
         starts_with_nonfill = ord(symbolic_centered[0]) != ord(fillch)
-        with NoTracing():
-            assert not space.is_possible(starts_with_nonfill.var)
+        assert not space.is_possible(starts_with_nonfill)
 
 
 def test_str_map_chars() -> None:
@@ -1195,12 +1277,12 @@ def test_str___add___method() -> None:
 def test_str_bool():
     with standalone_statespace as space, NoTracing():
         a = LazyIntSymbolicStr("a")
-        space.add(a.__len__().var > 0)
         with ResumedTracing():
+            space.add(a.__len__() > 0)
             assert bool(a)
-        # Can we retain our symbolic state after forcing a positive truthiness?:
-        assert space.is_possible((a == "this").var)
-        assert space.is_possible((a == "that").var)
+            # Can we retain our symbolic state after forcing a positive truthiness?:
+            assert space.is_possible(a == "this")
+            assert space.is_possible(a == "that")
 
 
 def test_str_eq():
@@ -1223,6 +1305,18 @@ def test_str_filter():
         assert ret == [string]
 
 
+def test_str_filter_with_none():
+    with standalone_statespace as space, NoTracing():
+        string = LazyIntSymbolicStr([ord("a")])
+        truthyint = proxy_for_type(int, "truthyint")
+        falseyint = proxy_for_type(int, "falseyint")
+        with ResumedTracing():
+            space.add(truthyint == 10)
+            space.add(falseyint == 0)
+            ret = deep_realize(list(filter(None, [falseyint, 42, 0, truthyint])))
+        assert ret == [42, 10]
+
+
 def test_str_find() -> None:
     with standalone_statespace, NoTracing():
         string = LazyIntSymbolicStr(list(map(ord, "aabc")))
@@ -1250,20 +1344,18 @@ def test_str_format_basic():
     with standalone_statespace as space:
         with NoTracing():
             s = LazyIntSymbolicStr("s")
-            space.add(s.__len__().var == 1)
-        assert space.is_possible((s == "z").var)
-        assert space.is_possible((ord("a{0}c".format(s)[1]) == ord("b")).var)
+        space.add(s.__len__() == 1)
+        assert space.is_possible(s == "z")
+        assert space.is_possible(ord("a{0}c".format(s)[1]) == ord("b"))
 
 
 def test_str_format_map():
     with standalone_statespace as space:
         with NoTracing():
             s = LazyIntSymbolicStr("s")
-            space.add(s.__len__().var == 1)
-        assert space.is_possible((s == "z").var)
-        assert space.is_possible(
-            (ord("a{foo}c".format_map({"foo": s})[1]) == ord("b")).var
-        )
+        space.add(s.__len__() == 1)
+        assert space.is_possible(s == "z")
+        assert space.is_possible(ord("a{foo}c".format_map({"foo": s})[1]) == ord("b"))
 
 
 def test_str_rfind() -> None:
@@ -1406,6 +1498,23 @@ def test_tuple___len___method():
     check_states(f, POST_FAIL)
 
 
+def test_tuple___repr__symbolic_in_concrete(space) -> None:
+    x = proxy_for_type(int, "x")
+    with ResumedTracing():
+        space.add(x == 4)  # type: ignore
+        container = (x, x)
+        assert repr(container) == "(4, 4)"
+
+
+def test_tuple___repr__symbolic_in_concrete_namedtuple(space) -> None:
+    NamedTupleClass = collections.namedtuple("NamedTupleClass", ["target"])
+    x = proxy_for_type(int, "x")
+    with ResumedTracing():
+        space.add(x == 4)  # type: ignore
+        container = NamedTupleClass(target=x)
+        assert repr(container) == "NamedTupleClass(target=4)"
+
+
 def test_tuple_range_intersection_fail() -> None:
     def f(a: Tuple[int, int], b: Tuple[int, int]) -> Optional[Tuple[int, int]]:
         """
@@ -1460,6 +1569,13 @@ def test_tuple_runtime_type() -> None:
     check_states(f, POST_FAIL)
 
 
+def test_empty_tuple(space) -> None:
+    t = proxy_for_type(Tuple[()], "t")
+    with ResumedTracing():
+        assert type(t) is tuple
+        assert t == ()
+
+
 def test_tuple_isinstance_check() -> None:
     def f(uniform_tuple: Tuple[List, ...], basic_tuple: tuple) -> Tuple[bool, bool]:
         """post: _ == (True, True)"""
@@ -1492,6 +1608,19 @@ def test_range___reversed___method() -> None:
     check_states(f, POST_FAIL)
 
 
+def test_range_slicing(space) -> None:
+    assert list(range(3, 40, 2)[5::-1]) == [13, 11, 9, 7, 5, 3]
+    # Repeat this with a symbolic range:
+    rng = proxy_for_type(range, "rng")
+    newstart = proxy_for_type(int, "newstart")
+    with ResumedTracing():
+        space.add(rng.start == 3)  # type: ignore
+        space.add(rng.stop == 40)  # type: ignore
+        space.add(rng.step == 2)  # type: ignore
+        space.add(newstart == 5)  # type: ignore
+        assert list(rng[newstart::-1]) == [13, 11, 9, 7, 5, 3]
+
+
 def test_range_realization(space) -> None:
     rng = proxy_for_type(range, "rng")
     assert isinstance(rng, SymbolicRange)
@@ -1541,6 +1670,14 @@ def test_list___add___method() -> None:
     check_states(f, POST_FAIL)
 
 
+def test_list___repr___symbolic_in_concrete(space) -> None:
+    x = proxy_for_type(int, "x")
+    with ResumedTracing():
+        space.add(x == 4)  # type: ignore
+        continer = [x]
+        assert f"{continer=}" == "continer=[4]"
+
+
 def test_list_doubling_ok() -> None:
     def f(a: List[int]) -> List[int]:
         """
@@ -1924,8 +2061,8 @@ def test_list_shallow_realization():
     with standalone_statespace as space:
         nums = proxy_for_type(List[int], "nums")
         numslen = len(nums)
+        space.add(numslen == 1)
         with NoTracing():
-            space.add(numslen.var == 1)
             realized = realize(nums)
             assert type(realized) is list
             assert len(realized) == 1
@@ -1934,15 +2071,16 @@ def test_list_shallow_realization():
 
 def test_list_concrete_with_symbolic_slice(space):
     idx = proxy_for_type(int, "i")
-    space.add(1 <= idx.var)
-    space.add(idx.var <= 3)
     with ResumedTracing():
+        space.add(1 <= idx)
+        space.add(idx <= 3)
         prefix = [0, 1, 2, 3][:idx]
         prefixlen = len(prefix)
     assert isinstance(prefix, CrossHairValue)
     assert isinstance(prefixlen, CrossHairValue)
-    assert space.is_possible(prefixlen.var == 1)
-    assert space.is_possible(prefixlen.var == 3)
+    with ResumedTracing():
+        assert space.is_possible(prefixlen == 1)
+        assert space.is_possible(prefixlen == 3)
 
 
 def test_list_copy(space):
@@ -1959,6 +2097,7 @@ def test_list_copy_compare_without_forking(space):
     lst = proxy_for_type(List[int], "lst")
     with ResumedTracing():
         lst2 = copy.deepcopy(lst)
+    assert lst.inner is not lst2.inner
     assert type(lst2) is SymbolicList
     assert lst.inner.var is lst2.inner.var
     with ResumedTracing():
@@ -2001,6 +2140,19 @@ def test_dict___iter___ok() -> None:
     check_states(f, CONFIRMED)
 
 
+def test_dict___or___method():
+    with standalone_statespace as space:
+        d = proxy_for_type(Dict[int, int], "d")
+        space.add(len(d) == 0)
+        with pytest.raises(TypeError):
+            d | set()
+        if sys.version_info >= (3, 9):
+            assert d | {1: 2} == {1: 2}
+        else:
+            with pytest.raises(TypeError):
+                d | {1: 2}
+
+
 @pytest.mark.demo("yellow")
 def test_dict___delitem___method() -> None:
     def f(a: Dict[str, int]) -> None:
@@ -2088,6 +2240,21 @@ def test_dict___getitem___implicit_conversion_for_keys_fail() -> None:
     check_states(f, POST_FAIL)
 
 
+def test_dict__items__works_with_symbolic_self(space) -> None:
+    x = proxy_for_type(Dict[int, int], "x")
+    with ResumedTracing():
+        x[42] = 42
+        assert (42, 42) in list(dict.items(x))
+
+
+def test_dict___repr___symbolic_in_concrete(space) -> None:
+    x = proxy_for_type(int, "x")
+    with ResumedTracing():
+        space.add(x == 4)  # type: ignore
+        container = {x: x}
+        assert repr(container) == "{4: 4}"
+
+
 @pytest.mark.demo("yellow")
 def test_dict___setitem___method() -> None:
     def f(a: Dict[int, int], k: int, v: int) -> None:
@@ -2202,6 +2369,26 @@ def test_dict_setdefault_float_int_comparison() -> None:
     check_states(f, CONFIRMED)
 
 
+def test_dict_construction_from_generator(space):
+    with ResumedTracing():
+        d = dict((k, v) for k, v in [(1, 2), (3, 4)])
+        assert d == {1: 2, 3: 4}
+
+
+def test_dict_construction_copies_dict(space):
+    with ResumedTracing():
+        source = {1: 2, 3: 4}
+        copy = dict(source)
+        source[5] = 6
+        assert copy == {1: 2, 3: 4}
+
+
+def test_dict_construction_with_duplicate_keys(space):
+    with ResumedTracing():
+        d = dict([(1, 2), (3, 4), (1, 10), (5, 6), (3, 10)])
+        assert d == {1: 10, 3: 10, 5: 6}
+
+
 def test_dict_over_objects() -> None:
     def f(a: Dict[object, object]) -> int:
         """
@@ -2344,9 +2531,10 @@ def test_dict_inside_lists() -> None:
     check_states(f, POST_FAIL)
 
 
+@pytest.mark.skip(
+    reason="flakey; seems to get stuck sometimes despite timeout extensions"
+)
 def test_dict_inside_lists_with_identity() -> None:
-    # NOTE: the message is a little confusing because repr()
-    # hides the fact that the identity of the lists is the same.
     def f(dicts: List[Dict[int, int]]):
         """
         Removes duplicate keys.
@@ -2354,6 +2542,7 @@ def test_dict_inside_lists_with_identity() -> None:
         pre:  len(dicts[0]) == 1
         post: len(dicts[0]) == 1
         """
+        # crosshair: max_uninteresting_iterations=50
         seen: Set[int] = set()
         for d in dicts:
             for k in d.keys():
@@ -2403,6 +2592,7 @@ def test_dict_alternate_mapping_types() -> None:
 def test_dict_untyped_access():
     def f(d: dict, k: int) -> dict:
         """
+        pre: len(d) == 2  # (just to bound the search space a little)
         pre: 42 in d
         post: 42 in __return__
         raises: KeyError
@@ -2410,11 +2600,7 @@ def test_dict_untyped_access():
         del d[k]
         return d
 
-    # TODO: profile / optimize
-    check_states(
-        f,
-        MessageType.POST_FAIL,
-    )
+    check_states(f, MessageType.POST_FAIL)
 
 
 # NOTE: TypedDict appeared earlier than 3.9, but was not runtime-detectable until then
@@ -2539,6 +2725,16 @@ def test_set_isinstance_check() -> None:
     check_states(f, CONFIRMED)
 
 
+def test_frozenset___eq__(space):
+    fs = proxy_for_type(Set[FrozenSet[int]], "fs")
+    with ResumedTracing():
+        space.add(len(fs) == 1)
+        space.add(len(next(iter(fs))) == 0)
+        linearset = set([frozenset()])
+        iseq = realize(fs == linearset)
+    assert iseq is True
+
+
 def test_set___eq__() -> None:
     def f(a: Set[FrozenSet[int]]) -> object:
         """
@@ -2554,6 +2750,31 @@ def test_set___eq__() -> None:
     )
 
 
+def test_frozenset___repr__symbolic_in_concrete(space) -> None:
+    x = proxy_for_type(int, "x")
+    with ResumedTracing():
+        space.add(x == 4)  # type: ignore
+        container = frozenset([x])
+        assert repr(container) == "frozenset({4})"
+
+
+def test_set___repr__symbolic_in_concrete(space) -> None:
+    x = proxy_for_type(int, "x")
+    with ResumedTracing():
+        space.add(x == 4)  # type: ignore
+        container = {x}
+        assert repr(container) == "{4}"
+
+
+def test_set_copy(space):
+    x = proxy_for_type(Set[int], "x")
+    with ResumedTracing():
+        space.add(len(x) == 1)
+        (y, z) = copy.deepcopy((x, x))
+        assert not space.is_possible(len(y) != 1)
+        assert not space.is_possible(list(x)[0] != list(y)[0])
+
+
 def test_set_no_duplicates() -> None:
     def f(s: Set[int]) -> int:
         """
@@ -2568,6 +2789,19 @@ def test_set_no_duplicates() -> None:
     check_states(f, CONFIRMED)
 
 
+def test_set_additions_to_concrete() -> None:
+    class F:
+        set_type = set
+
+    def f(self: F, x: int) -> Set[int]:
+        """
+        post: _ != set([1234])
+        """
+        return self.set_type([x, x])
+
+    check_states(f, POST_FAIL)
+
+
 def test_frozenset_realize():
     with standalone_statespace as space:
         with NoTracing():
@@ -2579,9 +2813,9 @@ def test_frozenset_realize():
 
 def test_frozenset_covariance():
     with standalone_statespace as space:
+        frozen_set = proxy_for_type(FrozenSet[AbstractBase], "x")
+        space.add(frozen_set.__len__() == 1)
         with NoTracing():
-            frozen_set = proxy_for_type(FrozenSet[AbstractBase], "x")
-            space.add(frozen_set.__len__().var == 1)
             assert isinstance(next(iter(frozen_set)), ConcreteSubclass)
 
 
@@ -2591,9 +2825,9 @@ def test_set_invariance():
     # assigned to a superclass container.
     # Therefore, CrossHair should happily create set contents using subclasses:
     with standalone_statespace as space:
+        mutable_set = proxy_for_type(Set[AbstractBase], "x")
+        space.add(mutable_set.__len__() == 1)
         with NoTracing():
-            mutable_set = proxy_for_type(Set[AbstractBase], "x")
-            space.add(mutable_set.__len__().var == 1)
             assert isinstance(next(iter(mutable_set)), ConcreteSubclass)
 
 
@@ -2611,14 +2845,43 @@ def test_set_iter_partial():
     with standalone_statespace as space:
         with NoTracing():
             x = proxy_for_type(Set[int], "x")
-            space.add(x.__len__().var == 2)
+        space.add(x.__len__() == 2)
         itr = iter(x)
         first = next(itr)
         # leave the iterator incomplete; looking for generator + context mgr problems
-    return
 
 
-class ProtocolsTest(unittest.TestCase):
+def test_set_containment_check_without_iteration():
+    with standalone_statespace as space:
+        x = proxy_for_type(FrozenSet[int], "x")
+        with ResumedTracing():
+            space.add(len(x) == 1)
+            assert space.is_possible(x._is_consistent())
+            space.add(x.__contains__(42))
+            assert space.is_possible(x._is_consistent())
+            space.add(x.__contains__(10))
+            assert not space.is_possible(x._is_consistent())
+
+
+def test_set_independence(space):
+    with ResumedTracing():
+        s = set()
+        copy = set(s)
+        s |= s
+        s != copy
+
+
+def test_frozenset___or__(space):
+    x = proxy_for_type(int, "x")
+    y = proxy_for_type(int, "y")
+    with ResumedTracing():
+        space.add(x != y)
+        s1 = frozenset([x])
+        s2 = frozenset([y])
+        assert len(s1 | s2) == 2
+
+
+class TestProtocols:
     # TODO: move most of this into a collectionslib_test.py file
     def test_hashable_values_fail(self) -> None:
         def f(b: bool, i: int, t: Tuple[str, ...]) -> int:
@@ -2650,7 +2913,10 @@ class ProtocolsTest(unittest.TestCase):
             # c: SupportsComplex,  # TODO: symbolic complex not yet really working
             b: SupportsBytes,
         ) -> float:
-            """post: _.real <= 0"""
+            """
+            pre: math.isfinite(f) and math.isfinite(r)
+            post: _.real <= 0
+            """
             return abs(a) + float(f) + int(i) + round(r) + len(bytes(b))
             # + complex(c)
 
@@ -2774,6 +3040,12 @@ def test_object___eq__() -> None:
     check_states(f, POST_FAIL)
 
 
+def test_object_get_type_hints(space: StateSpace) -> None:
+    typ = proxy_for_type(type, "typ")
+    with ResumedTracing():
+        get_type_hints(typ)
+
+
 def test_issubclass_abc():
     with standalone_statespace as space:
         with NoTracing():
@@ -2799,6 +3071,31 @@ def test_object_with_comparison():
     check_states(f, POST_FAIL)
 
 
+def test_object_addition():
+    def f(a, b):
+        """post: True"""
+        return a + b
+
+    actual, expected = check_messages(
+        analyze_function(f),
+        state=MessageType.EXEC_ERR,
+        # We check the message here, because we used to produce an
+        # (incorrect) counterexample of f('', '')
+        # See https://github.com/pschanely/CrossHair/issues/235
+        message="TypeError:  when calling f('', 0)",
+    )
+    assert actual == expected
+
+
+def test_invoke_integer():
+    # https://github.com/pschanely/CrossHair/issues/236
+    def f(a: int):
+        """post: True"""
+        return a(1)  # type: ignore
+
+    check_states(f, EXEC_ERR)
+
+
 def test_callable_zero_args() -> None:
     def f(size: int, initializer: Callable[[], int]) -> Tuple[int, ...]:
         """
@@ -2894,9 +3191,9 @@ def test_all():
 
     with standalone_statespace as space:
         true_bool = proxy_for_type(bool, "true_bool")
-        space.add(true_bool.var)
+        space.add(true_bool)
         nonempty_list = proxy_for_type(List[object], "nonempty_list")
-        space.add(len(nonempty_list).var == 1)
+        space.add(len(nonempty_list) == 1)
         arbitrary_bool = proxy_for_type(bool, "arbitrary_bool")
         boom = ExplodingBool()
 
@@ -2905,8 +3202,8 @@ def test_all():
         sym_false = all([True, true_bool, nonempty_list, arbitrary_bool])
         with NoTracing():
             assert isinstance(sym_false, SymbolicBool)
-            assert space.is_possible(sym_false.var)
             assert space.is_possible(z3.Not(sym_false.var))
+        assert space.is_possible(sym_false)
 
         with pytest.raises(ExplodingBool):
             all([true_bool, True, nonempty_list, boom])
@@ -2923,7 +3220,7 @@ def test_any():
         false_bool = proxy_for_type(bool, "false_bool")
         space.add(z3.Not(false_bool.var))
         empty_list = proxy_for_type(List[object], "empty_list")
-        space.add(len(empty_list).var == 0)
+        space.add(len(empty_list) == 0)
         arbitrary_bool = proxy_for_type(bool, "arbitrary_bool")
         # arbitrary_list = proxy_for_type(List[object], "arbitrary_list")
         # TODO: in theory, we should be able to keep `bool(arbitrary_list)`
@@ -2936,8 +3233,8 @@ def test_any():
         sym_false = any([False, false_bool, empty_list, arbitrary_bool])
         with NoTracing():
             assert isinstance(sym_false, SymbolicBool)
-            assert space.is_possible(sym_false.var)
             assert space.is_possible(z3.Not(sym_false.var))
+        assert space.is_possible(sym_false)
 
         with pytest.raises(ExplodingBool):
             any([false_bool, False, empty_list, boom])
@@ -3085,6 +3382,10 @@ def test_bytes_roundtrip_array_as_symbolic():
             assert new_bytes.inner is orig_bytes.inner
 
 
+@pytest.mark.skipif(
+    sys.version_info >= (3, 13),
+    reason="Need to intercept UnicodeDecodeError.str in 3.13+",
+)
 @pytest.mark.demo
 def test_bytes_decode_method():
     def f(b: bytes) -> str:
@@ -3141,13 +3442,13 @@ def test_extend_concrete_bytearray():
         xyz = proxy_for_type(bytearray, "xyz")
         b.extend(xyz)
         assert not space.is_possible(b[0] != ord("a"))
-        assert space.is_possible(len(b).var > 3)
+        assert space.is_possible(len(b) > 3)
 
 
 def test_bytearray_slice():
     with standalone_statespace as space:
         xyz = proxy_for_type(bytearray, "xyz")
-        space.add(xyz.__len__().var == 3)
+        space.add(xyz.__len__() == 3)
         assert type(xyz[1:]) is bytearray
 
 
@@ -3156,9 +3457,8 @@ def test_memoryview_compare():
         mv1 = proxy_for_type(memoryview, "mv1")
         mv2 = proxy_for_type(memoryview, "mv2")
         len1, len2 = len(mv1), len(mv2)
-        with NoTracing():
-            space.add(len1.var == 0)
-            space.add(len2.var == 0)
+        space.add(len1 == 0)
+        space.add(len2 == 0)
         views_equal = mv1 == mv2
         with NoTracing():
             assert views_equal is True
@@ -3168,7 +3468,7 @@ def test_memoryview_cast():
     """post: _"""
     with standalone_statespace as space:
         val = proxy_for_type(int, "val")
-        space.add(val.var == 254)
+        space.add(val == 254)
         mv = memoryview(bytearray([val]))
         assert mv.cast("b")[0] == -2
 
@@ -3177,7 +3477,7 @@ def test_memoryview_toreadonly():
     """post: _"""
     with standalone_statespace as space:
         mv = proxy_for_type(memoryview, "mv")
-        space.add(mv.__len__().var == 1)
+        space.add(mv.__len__() == 1)
         mv2 = mv.toreadonly()
         mv[0] = 12
         assert mv2[0] == 12
@@ -3189,7 +3489,7 @@ def test_memoryview_properties():
     """post: _"""
     with standalone_statespace as space:
         symbolic_mv = proxy_for_type(memoryview, "symbolic_mv")
-        space.add(symbolic_mv.__len__().var == 1)
+        space.add(symbolic_mv.__len__() == 1)
         concrete_mv = memoryview(bytearray(b"a"))
         assert symbolic_mv.contiguous == concrete_mv.contiguous
         assert symbolic_mv.c_contiguous == concrete_mv.c_contiguous
@@ -3204,46 +3504,47 @@ def test_memoryview_properties():
         assert symbolic_mv.suboffsets == concrete_mv.suboffsets
 
 
-def test_chr():
-    with standalone_statespace as space:
-        i = proxy_for_type(int, "i")
-        space.add(z3.And(10 <= i.var, i.var < 256))
+def test_chr(space):
+    i = proxy_for_type(int, "i")
+    with ResumedTracing():
+        space.add(10 <= i)
+        space.add(i < 256)
         c = chr(i)
-        assert space.is_possible(c._codepoints[0].var == ord("a"))
-        assert not space.is_possible(c._codepoints[0].var == 0)
+        assert space.is_possible(c._codepoints[0] == ord("a"))
+        assert not space.is_possible(c._codepoints[0] == 0)
 
 
-def test_ord():
-    with standalone_statespace as space:
-        s = proxy_for_type(str, "s")
-        space.add(len(s).var == 1)
+def test_ord(space):
+    s = proxy_for_type(str, "s")
+    with ResumedTracing():
+        space.add(len(s) == 1)
         i = ord(s)
-        assert space.is_possible(i.var == 42)
-        assert not space.is_possible(i.var > sys.maxunicode)
+        assert space.is_possible(i == 42)
+        assert not space.is_possible(i > sys.maxunicode)
 
 
-def test_unicode_support():
-    with standalone_statespace as space:
-        s = proxy_for_type(str, "s")
-        space.add(len(s).var == 1)
-        assert space.is_possible((s == "a").var)
-        assert space.is_possible((s == "\u1234").var)
+def test_unicode_support(space):
+    s = proxy_for_type(str, "s")
+    with ResumedTracing():
+        space.add(len(s) == 1)
+        assert space.is_possible(s == "a")
+        assert space.is_possible(s == "\u1234")
 
 
 @pytest.mark.parametrize("concrete_x", (25, 15, 6, -4, -15, -25))
-def test_int_round(concrete_x):
-    with standalone_statespace as space:
-        concrete_ret = round(concrete_x, -1)
-        x = proxy_for_type(int, "x")
-        d = proxy_for_type(int, "d")
-        space.add(x.var == concrete_x)
-        space.add(d.var == -1)
-        assert not space.is_possible((round(x, d) != concrete_ret).var)
+def test_int_round(concrete_x, space):
+    concrete_ret = round(concrete_x, -1)
+    x = proxy_for_type(int, "x")
+    d = proxy_for_type(int, "d")
+    with ResumedTracing():
+        space.add(x == concrete_x)
+        space.add(d == -1)
+        assert not space.is_possible((round(x, d) != concrete_ret))
 
 
 class ExplodingValue:
     def __getattribute__(self, name):
-        raise CrosshairInternal
+        raise CrossHairInternal
 
 
 @dataclasses.dataclass
@@ -3265,6 +3566,58 @@ def test_class_with_special_member_names():
     check_states(f, CONFIRMED)
 
 
+@pytest.mark.parametrize("type2", (list, set, frozenset, tuple))
+@pytest.mark.parametrize(
+    "type1",
+    (
+        list,
+        set,
+        frozenset,
+        tuple,
+        array,
+    ),
+)
+def test_container_crosstype_equality(space, type1, type2):
+    sym_obj1 = proxy_for_type(type1, "obj1")
+    obj2 = type2()
+    with ResumedTracing():
+        space.add(len(sym_obj1) == 0)
+        obj1 = deep_realize(sym_obj1)
+        sym_eq_on_left = realize(sym_obj1 == obj2)
+        sym_eq_on_right = realize(obj2 == sym_obj1)
+    actually_equal = obj1 == obj2
+    assert actually_equal == sym_eq_on_left
+    assert actually_equal == sym_eq_on_right
+
+
+@pytest.mark.parametrize("type2", (str, bytes, tuple))
+@pytest.mark.parametrize("type1", (str, bytes, tuple, memoryview))
+def test_stringlike_crosstype_equality(space, type1, type2):
+    sym_obj1 = proxy_for_type(type1, "obj1")
+    obj2 = type2()
+    with ResumedTracing():
+        space.add(len(sym_obj1) == 0)
+        obj1 = deep_realize(sym_obj1)
+        sym_eq_on_left = realize(sym_obj1 == obj2)
+        sym_eq_on_right = realize(obj2 == sym_obj1)
+    actually_equal = obj1 == obj2
+    assert actually_equal == sym_eq_on_left
+    assert actually_equal == sym_eq_on_right
+
+
+@pytest.mark.parametrize("typ", (List[int],))
+def test_deep_realization(space, typ):
+    symbolic = proxy_for_type(typ, "symbolic")
+    with ResumedTracing():
+        space.add(len(symbolic) == 3)
+    origin = origin_of(typ)
+    assert type(symbolic) != origin
+    concrete = deep_realize(symbolic)
+    assert type(concrete) == origin
+    with ResumedTracing():
+        assert concrete == symbolic
+
+
 def TODO_test_float_precision_issues(a, b):
     # Does not yet work: floating point is modeled with infinite precision at the moment
     def f(a: int, b: int):
@@ -3286,8 +3639,8 @@ def TODO_test_int_mod_float():
         y = proxy_for_type(float, "y")
         modval = x % y
         with NoTracing():
-            assert type(modval) == SymbolicFloat
-            assert space.is_possible(modval.var == 12.12)
+            assert type(modval) == RealBasedSymbolicFloat
+        assert space.is_possible(modval == 12.12)
 
 
 def TODO_test_can_generate_integral():
@@ -3303,9 +3656,3 @@ def TODO_test_deepcopy_independence():
         with NoTracing():
             assert ls[0] is not lscopy[0]
         # Next try mutation on one and test the other...
-
-
-if __name__ == "__main__":
-    if ("-v" in sys.argv) or ("--verbose" in sys.argv):
-        set_debug(True)
-    unittest.main()
