@@ -1,21 +1,25 @@
 import collections
+import sys
 from typing import (
     Any,
     Callable,
     Dict,
     Generic,
     Iterable,
+    KeysView,
     List,
     Optional,
     Set,
     Tuple,
     TypeVar,
+    Union,
+    ValuesView,
 )
 
 from crosshair import register_type
-from crosshair.core import CrossHairValue, realize
+from crosshair.core import realize
 from crosshair.tracers import NoTracing, ResumedTracing
-from crosshair.util import is_iterable
+from crosshair.util import CrossHairValue, is_iterable
 
 T = TypeVar("T")
 
@@ -203,11 +207,29 @@ def make_registrations():
     register_type(collections.abc.Mapping, lambda p, kt=Any, vt=Any: p(Dict[kt, vt]))  # type: ignore
     register_type(collections.abc.MutableMapping, lambda p, kt=Any, vt=Any: p(Dict[kt, vt]))  # type: ignore
     register_type(collections.OrderedDict, lambda p, kt=Any, vt=Any: collections.OrderedDict(p(Dict[kt, vt])))  # type: ignore
+    # TODO: This implementation of Counter probably over-realizes the symbolic map it is given:
     register_type(collections.Counter, lambda p, t=Any: collections.Counter(p(Dict[t, int])))  # type: ignore
     # TODO: MappingView is missing
-    register_type(collections.abc.ItemsView, lambda p, kt=Any, vt=Any: p(Set[Tuple[kt, vt]]))  # type: ignore
-    register_type(collections.abc.KeysView, lambda p, t=Any: p(Set[t]))  # type: ignore
-    register_type(collections.abc.ValuesView, lambda p, t=Any: p(List[t]))  # type: ignore
+    register_type(
+        collections.abc.ItemsView,
+        lambda p, k=Any, v=Any: p(Dict.__getitem__((k, v))).items(),
+    )
+    register_type(
+        collections.abc.MappingView,
+        lambda p, t=Any: p(
+            Union[
+                KeysView.__getitem__((t,)),
+                ValuesView.__getitem__((t,)),
+            ]
+        ),
+    )
+    register_type(
+        collections.abc.KeysView, lambda p, t=Any: p(Dict.__getitem__((t, Any))).keys()
+    )
+    register_type(
+        collections.abc.ValuesView,
+        lambda p, t=Any: p(Dict.__getitem__((Any, t))).values(),
+    )
 
     register_type(collections.abc.Container, lambda p, t=Any: p(Tuple[t, ...]))
     register_type(collections.abc.Collection, lambda p, t=Any: p(Tuple[t, ...]))
@@ -224,5 +246,8 @@ def make_registrations():
 
     register_type(collections.abc.MutableSet, lambda p, t=Any: p(Set[t]))  # type: ignore
 
-    register_type(collections.abc.ByteString, lambda p: p(bytes))
+    if sys.version_info < (3, 14):
+        register_type(collections.abc.ByteString, lambda p: p(bytes))
+    if sys.version_info >= (3, 12):
+        register_type(collections.abc.Buffer, lambda p: p(bytes))
     register_type(collections.abc.Hashable, lambda p: p(int))
