@@ -36,7 +36,7 @@ from crosshair.core_and_libs import (
     installed_plugins,
     run_checkables,
 )
-from crosshair.diff_behavior import diff_behavior
+from crosshair.diff_behavior import ExceptionEquivalenceType, diff_behavior
 from crosshair.fnutil import (
     FUNCTIONINFO_DESCRIPTOR_TYPES,
     FunctionInfo,
@@ -60,10 +60,9 @@ from crosshair.path_cover import (
 from crosshair.path_search import OptimizationKind, path_search
 from crosshair.pure_importer import prefer_pure_python_imports
 from crosshair.register_contract import REGISTERED_CONTRACTS
-from crosshair.statespace import NotDeterministic, context_statespace
-from crosshair.tracers import NoTracing
 from crosshair.util import (
     ErrorDuringImport,
+    NotDeterministic,
     add_to_pypath,
     debug,
     format_boundargs,
@@ -86,14 +85,6 @@ def analysis_kind(argstr: str) -> Sequence[AnalysisKind]:
         ret = [AnalysisKind[part.strip()] for part in argstr.split(",")]
     except KeyError:
         raise ValueError
-    if AnalysisKind.hypothesis in ret:
-        try:
-            import hypothesis
-
-            if hypothesis.__version_info__ < (6, 0, 0):
-                raise Exception("CrossHair requires hypothesis version >= 6.0.0")
-        except ImportError as e:
-            raise Exception("Unable to import the hypothesis library") from e
     return ret
 
 
@@ -168,11 +159,11 @@ def command_line_parser() -> argparse.ArgumentParser:
         formatter_class=argparse.RawTextHelpFormatter,
         description=textwrap.dedent(
             """\
-        The search command finds arguments for a function that causes it to complete without
-        error.
+        The search command finds arguments for a function that causes it to
+        complete without error.
 
-        Results (if any) are written to stdout in the form of a repr'd dictionary, mapping
-        argument names to values.
+        Results (if any) are written to stdout in the form of a repr'd
+        dictionary, mapping argument names to values.
         """
         ),
     )
@@ -191,13 +182,13 @@ def command_line_parser() -> argparse.ArgumentParser:
         help=textwrap.dedent(
             """\
         Controls what kind of arguments are produced.
-        Optimization effectiveness will vary wildly depnding on the nature of the
-        function.
-            simplify     : [default] Attempt to minimize the size (in characters) of the
-                           arguments.
+        Optimization effectiveness will vary wildly depnding on the nature of
+        the function.
+            simplify     : [default] Attempt to minimize the size
+                           (in characters) of the arguments.
             none         : Output the first set of arguments found.
-            minimize_int : Attempt to minimize an integer returned by the function.
-                           Negative return values are ignored.
+            minimize_int : Attempt to minimize an integer returned by the
+                           function. Negative return values are ignored.
         """
         ),
     )
@@ -217,11 +208,11 @@ def command_line_parser() -> argparse.ArgumentParser:
         type=str,
         help=textwrap.dedent(
             """\
-        The (fully-qualified) name of a function for formatting produced arguments.
-        If specified, crosshair will call this function instead of repr() when printing
-        arguments to stdout.
-        Your formatting function will be pased an `inspect.BoundArguments` instance.
-        It should return a string.
+        The (fully-qualified) name of a function for formatting produced
+        arguments. If specified, crosshair will call this function instead of
+        repr() when printing arguments to stdout.
+        Your formatting function will be pased an `inspect.BoundArguments`
+        instance. It should return a string.
         """
         ),
     )
@@ -274,14 +265,29 @@ def command_line_parser() -> argparse.ArgumentParser:
         type=str,
         help="second fully-qualified function to compare",
     )
+    diffbehavior_parser.add_argument(
+        "--exception_equivalence",
+        metavar="EXCEPTION_EQUIVALENCE",
+        type=ExceptionEquivalenceType,
+        default=ExceptionEquivalenceType.TYPE_AND_MESSAGE,
+        choices=ExceptionEquivalenceType.__members__.values(),
+        help=textwrap.dedent(
+            """\
+            Decide how to treat exceptions, while searching for a counter-example.
+            `ALL` treats all exceptions as equivalent,
+            `SAME_TYPE`, considers matches on the type.
+            `TYPE_AND_MESSAGE` matches for the same type and message.
+            """
+        ),
+    )
     cover_parser = subparsers.add_parser(
         "cover",
         formatter_class=argparse.RawTextHelpFormatter,
         help="Generate inputs for a function, attempting to exercise different code paths",
         description=textwrap.dedent(
             """\
-        Generates inputs to a function, hopefully getting good line, branch, and path
-        coverage.
+        Generates inputs to a function, hopefully getting good line, branch,
+        and path coverage.
         See https://crosshair.readthedocs.io/en/latest/cover.html
             """
         ),
@@ -310,9 +316,10 @@ def command_line_parser() -> argparse.ArgumentParser:
         help=textwrap.dedent(
             """\
         Determines how to output examples.
-            eval_expression     : [default] Output examples as expressions, suitable for
-                                  eval()
-            arg_dictionary      : Output arguments as repr'd, ordered dictionaries
+            eval_expression     : [default] Output examples as expressions,
+                                  suitable for eval()
+            arg_dictionary      : Output arguments as repr'd, ordered
+                                  dictionaries
             pytest              : Output examples as stub pytest tests
             argument_dictionary : Deprecated
         """
@@ -327,33 +334,35 @@ def command_line_parser() -> argparse.ArgumentParser:
         help=textwrap.dedent(
             """\
         Determines what kind of coverage to achieve.
-            opcode : [default] Cover as many opcodes of the function as possible.
-                     This is similar to "branch" coverage.
+            opcode : [default] Cover as many opcodes of the function as
+                     possible. This is similar to "branch" coverage.
             path   : Cover any possible execution path.
-                     There will usually be an infinite number of paths (e.g. loops are
-                     effectively unrolled). Use max_uninteresting_iterations and/or
-                     per_condition_timeout to bound results.
-                     Many path decisions are internal to CrossHair, so you may see more
-                     duplicative-ness in the output than you'd expect.
+                     There will usually be an infinite number of paths (e.g.
+                     loops are effectively unrolled). Use
+                     max_uninteresting_iterations and/or per_condition_timeout
+                     to bound results.
+                     Many path decisions are internal to CrossHair, so you may
+                     see more duplicative-ness in the output than you'd expect.
         """
         ),
     )
-    for subparser in (check_parser, cover_parser, search_parser):
+    for subparser in (check_parser, search_parser, diffbehavior_parser, cover_parser):
         subparser.add_argument(
             "--max_uninteresting_iterations",
             type=int,
             help=textwrap.dedent(
                 """\
-            Maximum number of consequitive iterations to run without making
+            Maximum number of consecutive iterations to run without making
             significant progress in exploring the codebase.
+            (by default, 5 iterations, unless --per_condition_timeout is set)
 
-            This option can be useful than --per_condition_timeout
+            This option can be more useful than --per_condition_timeout
             because the amount of time invested will scale with the complexity
             of the code under analysis.
 
             Use a small integer (3-5) for fast but weak analysis.
-            Values in the hundreds or thousands may be appropriate if you intend to
-            run CrossHair for hours.
+            Values in the hundreds or thousands may be appropriate if you
+            intend to run CrossHair for hours.
             """
             ),
         )
@@ -366,11 +375,14 @@ def command_line_parser() -> argparse.ArgumentParser:
             help=textwrap.dedent(
                 """\
             Maximum seconds to spend checking one execution path.
-            If unspecified, CrossHair will timeout each path:
-            1. At the square root of `--per_condition_timeout`, if speficied.
-            2. Otherwise, at a number of seconds equal to
-               `--max_uninteresting_iterations`, if specified.
-            3. Otherwise, there will be no per-path timeout.
+            If unspecified:
+            1. CrossHair will timeout each path at the square root of
+               `--per_condition_timeout`, if specified.
+            3. Otherwise, it will timeout each path at a number of seconds
+               equal to `--max_uninteresting_iterations`, unless it is
+               explicitly set to zero.
+               (NOTE: `--max_uninteresting_iterations` is 5 by default)
+            2. Otherwise, it will not use any per-path timeout.
             """
             ),
         )
@@ -408,7 +420,6 @@ def command_line_parser() -> argparse.ArgumentParser:
                 PEP316     : check PEP316 contracts (docstring-based)
                 icontract  : check icontract contracts (decorator-based)
                 deal       : check deal contracts (decorator-based)
-                hypothesis : check hypothesis tests
             """
             ),
         )
@@ -437,7 +448,9 @@ def run_watch_loop(
             active_messages = {}
         else:
             time.sleep(0.1)
-            max_uninteresting_iterations *= 2
+            max_uninteresting_iterations = min(
+                max_uninteresting_iterations * 2, 100_000_000
+            )
         for curstats, messages in watcher.run_iteration(max_uninteresting_iterations):
             messages = [m for m in messages if m.state > MessageType.PRE_UNSAT]
             stats.update(curstats)
@@ -524,10 +537,10 @@ def messages_merged(
 _MOTD = [
     "Did I miss a counterexample? Let me know: https://github.com/pschanely/CrossHair/issues/new",
     "Help me be faster! Add to my benchmark suite: https://github.com/pschanely/crosshair-benchmark",
-    "Consider sharing your CrossHair experience on YouTube, Twitter, your blog, ... even TikTok.",
+    "Please consider sharing your CrossHair experience with others on social media.",
     "Questions? Ask at https://github.com/pschanely/CrossHair/discussions/new?category=q-a",
     "Consider signing up for CrossHair updates at https://pschanely.github.io",
-    # Use CrossHair? We'd like to reference your work here: ...
+    "Come say hello in the discord chat; we are friendly! https://discord.gg/rUeTaYTWbb",
 ]
 
 
@@ -683,10 +696,11 @@ def diffbehavior(
     (fn_name1, fn_name2) = (args.fn1, args.fn2)
     fn1 = checked_fn_load(fn_name1, stderr)
     fn2 = checked_fn_load(fn_name2, stderr)
+    exception_equivalence = args.exception_equivalence
     if fn1 is None or fn2 is None:
         return 2
     options.stats = Counter()
-    diffs = diff_behavior(fn1, fn2, options)
+    diffs = diff_behavior(fn1, fn2, options, exception_equivalence)
     debug("stats", options.stats)
     if isinstance(diffs, str):
         print(diffs, file=stderr)
@@ -931,6 +945,7 @@ def mypy_and_check(cmd_args: Optional[List[str]] = None) -> None:
         _mypy_out, mypy_err, mypy_ret = api.run(mypy_cmd_args)
         print(mypy_err, file=sys.stderr)
         if mypy_ret != 0:
+            print(_mypy_out, file=sys.stdout)
             sys.exit(mypy_ret)
     engage_auditwall()
     debug("Running crosshair with these args:", check_args)
