@@ -61,7 +61,12 @@ def call_check(
 def call_diffbehavior(fn1: str, fn2: str) -> Tuple[int, List[str]]:
     buf: io.StringIO = io.StringIO()
     errbuf: io.StringIO = io.StringIO()
-    retcode = diffbehavior(Namespace(fn1=fn1, fn2=fn2), DEFAULT_OPTIONS, buf, errbuf)
+    retcode = diffbehavior(
+        Namespace(fn1=fn1, fn2=fn2, exception_equivalence="type_and_message"),
+        DEFAULT_OPTIONS,
+        buf,
+        errbuf,
+    )
     lines = [
         ls for ls in buf.getvalue().split("\n") + errbuf.getvalue().split("\n") if ls
     ]
@@ -359,6 +364,35 @@ def test_check_circular_with_guard(root):
         assert retcode == 0
 
 
+def test_check_not_deterministic(root) -> None:
+    NOT_DETERMINISTIC_FOO = {
+        "foo.py": """
+_GLOBAL_THING = [42]
+
+def wonky_foo(i: int) -> int:
+    '''post: True'''
+    _GLOBAL_THING[0] += 1
+    if i > _GLOBAL_THING[0]:
+        pass
+    return True
+_GLOBAL_THING = [True]
+
+def regular_foo(i: int) -> int:
+    '''post: True'''
+    return i
+"""
+    }
+    simplefs(root, NOT_DETERMINISTIC_FOO)
+    with add_to_pypath(root):
+        retcode, lines, errlines = call_check([str(root / "foo.py")])
+        assert errlines == []
+        assert (
+            "NotDeterministic: Found a different execution paths after making the same decisions"
+            in lines[0]
+        )
+        assert retcode == 1
+
+
 def test_watch(root):
     # Just to make sure nothing explodes
     simplefs(root, SIMPLE_FOO)
@@ -456,7 +490,7 @@ def test_main_as_subprocess(tmp_path: Path):
     # the testing process.
     simplefs(tmp_path, SIMPLE_FOO)
     completion = subprocess.run(
-        [sys.executable, "-m", "crosshair", "check", str(tmp_path)],
+        [sys.executable, "-Werror", "-m", "crosshair", "check", str(tmp_path)],
         stdin=subprocess.DEVNULL,
         capture_output=True,
         text=True,
@@ -468,7 +502,7 @@ def test_main_as_subprocess(tmp_path: Path):
 
 def test_mypycrosshair_command():
     example_file = join(
-        split(__file__)[0], "examples", "icontract", "bugs_detected", "wrong_sign.py"
+        split(__file__)[0], "examples", "PEP316", "bugs_detected", "showcase.py"
     )
     completion = subprocess.run(
         [
@@ -482,7 +516,9 @@ def test_mypycrosshair_command():
         text=True,
     )
     assert completion.stderr.strip() == ""
-    assert completion.returncode == 1
+    if completion.returncode != 1:
+        print(completion.stdout)
+        assert False, f"Return code was {completion.returncode}"
 
 
 def test_describe_message():
