@@ -1,7 +1,9 @@
+import copy
 import sys
 
 import pytest
 
+from crosshair.core import deep_realize, realize
 from crosshair.simplestructs import (
     LazySetCombination,
     SequenceConcatenation,
@@ -35,6 +37,13 @@ def test_ShellMutableMap_popitem_ordering() -> None:
     assert SimpleDict([("c", "d"), ("a", "b")]).popitem() == ("a", "b")
 
 
+def test_ShellMutableMap_pop_with_default() -> None:
+    assert {"c": "d", "a": "b"}.popitem() == ("a", "b")
+    assert ShellMutableMap({}).pop("a", "x") == "x"
+    assert ShellMutableMap({"a": "b"}).pop("a", "x") == "b"
+    assert ShellMutableMap({"a": "b"}).pop("c", "x") == "x"
+
+
 def test_ShellMutableMap_poo() -> None:
     m = ShellMutableMap({2: 0})
     assert 0 == m.setdefault(2.0, {True: "0"})
@@ -78,6 +87,27 @@ def test_LazySetCombination_xor() -> None:
     assert 5 in s
 
 
+def test_ShellMutableDict_realize() -> None:
+    shell = ShellMutableMap({1: ShellMutableMap({2: 3})})
+    realized = realize(shell)
+    assert isinstance(realized, dict)
+    assert isinstance(realized[1], ShellMutableMap)
+
+
+def test_ShellMutableDict_deep_realize() -> None:
+    shell = ShellMutableMap({1: ShellMutableMap({2: 3})})
+    realized = deep_realize(shell)
+    assert isinstance(realized, dict)
+    assert isinstance(realized[1], dict)
+
+
+def test_ShellMutableSet_deepcopy() -> None:
+    ls = ["0", "1", "2", "3"]
+    shell = ShellMutableSet(ls)
+    # deep_realize(shell)
+    copy.deepcopy(shell)
+
+
 def test_ShellMutableSequence_slice_assignment() -> None:
     ls = ["0", "1", "2", "3"]
     shell = ShellMutableSequence(ls)
@@ -207,7 +237,7 @@ def test_ShellMutableSet_errors() -> None:
     with pytest.raises(KeyError):
         ShellMutableSet([]).pop()
     with pytest.raises(TypeError):
-        ShellMutableSet(4)
+        ShellMutableSet(4)  # type: ignore
 
 
 @pytest.mark.parametrize("start", [-3, -1, 0, 1, 3, None])
