@@ -10,38 +10,43 @@ from collections import Counter, defaultdict
 from dataclasses import dataclass
 from sys import _getframe
 from time import monotonic
+from traceback import extract_stack, format_tb
+from types import FrameType
 from typing import (
     Any,
     Callable,
     Dict,
     List,
     NewType,
+    NoReturn,
     Optional,
     Sequence,
     Set,
     Tuple,
     Type,
     TypeVar,
-    cast,
 )
 
 import z3  # type: ignore
 
 from crosshair import dynamic_typing
 from crosshair.condition_parser import ConditionExpr
+from crosshair.smtlib import parse_smtlib_literal
 from crosshair.tracers import NoTracing, ResumedTracing, is_tracing
 from crosshair.util import (
-    CrosshairInternal,
+    CROSSHAIR_EXTRA_ASSERTS,
+    CrossHairInternal,
     IgnoreAttempt,
+    NotDeterministic,
     PathTimeout,
     UnknownSatisfiability,
+    assert_tracing,
+    ch_stack,
     debug,
     in_debug,
     name_of_type,
-    origin_of,
-    test_stack,
 )
-from crosshair.z3util import z3Add, z3Not, z3PopNot
+from crosshair.z3util import z3Aassert, z3Not, z3PopNot
 
 
 @functools.total_ordering
@@ -148,6 +153,9 @@ SnapshotRef = NewType("SnapshotRef", int)
 
 def model_value_to_python(value: z3.ExprRef) -> object:
     if z3.is_real(value):
+        if isinstance(value, z3.AlgebraicNumRef):
+            # Force irrational values to be rational:
+            value = value.approx(precision=20)
         return float(value.as_fraction())
     elif z3.is_seq(value):
         ret = []
@@ -159,12 +167,14 @@ def model_value_to_python(value: z3.ExprRef) -> object:
         return ret
     elif z3.is_int(value):
         return value.as_long()
+    elif z3.is_fp(value):
+        return parse_smtlib_literal(value.sexpr())
     else:
         return ast.literal_eval(repr(value))
 
 
+@assert_tracing(False)
 def prefer_true(v: Any) -> bool:
-    assert not is_tracing()
     if not (hasattr(v, "var") and z3.is_bool(v.var)):
         with ResumedTracing():
             v = v.__bool__()
@@ -174,6 +184,22 @@ def prefer_true(v: Any) -> bool:
     return space.choose_possible(v.var, probability_true=1.0)
 
 
+def force_true(v: Any) -> None:
+    with NoTracing():
+        if not (hasattr(v, "var") and z3.is_bool(v.var)):
+            with ResumedTracing():
+                v = v.__bool__()
+            if not (hasattr(v, "var")):
+                raise CrossHairInternal(
+                    "Attempted to call assert_true on a non-symbolic"
+                )
+        space = context_statespace()
+        # TODO: we can improve this by making a new kind of (unary) assertion node
+        # that would not create these useless forks when the space is near exhaustion.
+        if not space.choose_possible(v.var, probability_true=1.0):
+            raise IgnoreAttempt
+
+
 class StateSpaceCounter(Counter):
     @property
     def iterations(self) -> int:
@@ -192,10 +218,6 @@ class StateSpaceCounter(Counter):
         return "{" + ", ".join(parts) + "}"
 
 
-class NotDeterministic(Exception):
-    pass
-
-
 class AbstractPathingOracle:
     def pre_path_hook(self, root: "RootNode") -> None:
         pass
@@ -223,12 +245,16 @@ class StateSpaceContext:
 
     def __enter__(self):
         prev = real_getattr(_THREAD_LOCALS, "space", None)
-        assert prev is None, "Already in a state space context"
-        _THREAD_LOCALS.space = self.space
+        if prev is not None:
+            raise CrossHairInternal("Already in a state space context")
+        space = self.space
+        _THREAD_LOCALS.space = space
+        space.mark_all_parent_frames()
 
     def __exit__(self, exc_type, exc_value, tb):
         prev = real_getattr(_THREAD_LOCALS, "space", None)
-        assert prev is self.space, "State space was altered in context"
+        if prev is not self.space:
+            raise CrossHairInternal("State space was altered in context")
         _THREAD_LOCALS.space = None
         return False
 
@@ -240,7 +266,7 @@ def optional_context_statespace() -> Optional["StateSpace"]:
 def context_statespace() -> "StateSpace":
     space = _THREAD_LOCALS.space
     if space is None:
-        raise CrosshairInternal
+        raise CrossHairInternal("Not in a statespace context")
     return space
 
 
@@ -291,7 +317,7 @@ class NodeStem(NodeLike):
         )
 
     def is_stem(self) -> bool:
-        return self.evolution is None
+        return self.evolution is None or self.evolution.is_stem()
 
     def grow_into(self, node: _N) -> _N:
         self.evolution = node
@@ -308,15 +334,12 @@ class NodeStem(NodeLike):
 
 
 class SearchTreeNode(NodeLike):
-    """
-    Represent a single decision point.
-
-    Abstract helper class for StateSpace.
-    """
+    """A node in the execution path tree."""
 
     stacktail: Tuple[str, ...] = ()
     result: CallAnalysis = CallAnalysis()
     exhausted: bool = False
+    iteration: Optional[int] = None
 
     def choose(
         self, space: "StateSpace", probability_true: Optional[float] = None
@@ -341,11 +364,15 @@ class SearchTreeNode(NodeLike):
         raise NotImplementedError
 
 
-def solver_is_sat(solver, *a) -> bool:
-    ret = solver.check(*a)
+def solver_is_sat(solver, *exprs) -> bool:
+    ret = solver.check(*exprs)
     if ret == z3.unknown:
-        solver.add(*a)
-        debug("Unknown satisfiability. Solver state follows:\n", solver.sexpr())
+        debug("Z3 Unknown satisfiability. Reason:", solver.reason_unknown())
+        if solver.reason_unknown() == "interrupted from keyboard":
+            raise KeyboardInterrupt
+        if exprs:
+            debug("While attempting to assert\n", *(e.sexpr() for e in exprs))
+        debug("Solver state follows:\n", solver.sexpr())
         raise UnknownSatisfiability
     return ret == z3.sat
 
@@ -419,6 +446,7 @@ class RootNode(SinglePathNode):
         from crosshair.pathing_oracle import CoveragePathingOracle  # circular import
 
         self.pathing_oracle = CoveragePathingOracle()
+        self.iteration = 0
 
 
 class DeatchedPathNode(SinglePathNode):
@@ -586,13 +614,11 @@ class WorstResultNode(RandomizedBinaryPathNode):
             if not solver_is_sat(solver, expr):
                 self.forced_path = False
         else:
-            # TODO: we still run into soundness issues on occasion, so I'd like to
-            # leave _PERFORM_EXTRA_SAT_CHECKS enabled a little longer:
-            _PERFORM_EXTRA_SAT_CHECKS = True
-            if _PERFORM_EXTRA_SAT_CHECKS and not solver_is_sat(solver, expr):
+            # We run into soundness issues on occasion:
+            if CROSSHAIR_EXTRA_ASSERTS and not solver_is_sat(solver, expr):
                 debug(" *** Reached impossible code path *** ")
                 debug("Current solver state:\n", str(solver))
-                raise CrosshairInternal("Reached impossible code path")
+                raise CrossHairInternal("Reached impossible code path")
             self.forced_path = True
         self.expr = expr
 
@@ -654,7 +680,7 @@ class ModelValueNode(WorstResultNode):
     def __init__(self, rand: random.Random, expr: z3.ExprRef, solver: z3.Solver):
         if not solver_is_sat(solver):
             debug("Solver unexpectedly unsat; solver state:", solver.sexpr())
-            raise CrosshairInternal("Unexpected unsat from solver")
+            raise CrossHairInternal("Unexpected unsat from solver")
 
         self.condition_value = solver.model().evaluate(expr, model_completion=True)
         self._stats_key = f"realize_{expr}" if z3.is_const(expr) else None
@@ -741,16 +767,30 @@ class StateSpace:
         self._random = search_root._random
         _, _, self._search_position = search_root.choose(self)
         self._deferred_assumptions = []
+        assert search_root.iteration is not None
+        search_root.iteration += 1
         search_root.pathing_oracle.pre_path_hook(search_root)
 
-    def add(self, expr: z3.ExprRef) -> None:
-        # debug('Committed to ', expr)
-        already_known = self._exprs_known.get(expr)
-        if already_known is None:
-            self.solver.add(expr)
-            self._exprs_known[expr] = True
-        elif already_known is not True:
-            raise CrosshairInternal
+    def add(self, expr) -> None:
+        with NoTracing():
+            if hasattr(expr, "var"):
+                expr = expr.var
+            elif not isinstance(expr, z3.ExprRef):
+                if type(expr) is bool:
+                    raise CrossHairInternal(
+                        "Attempted to assert a concrete boolean (look for unexpected realization)"
+                    )
+                raise CrossHairInternal(
+                    "Expected symbolic boolean, but supplied expression of type",
+                    name_of_type(type(expr)),
+                )
+            # debug('Committed to ', expr)
+            already_known = self._exprs_known.get(expr)
+            if already_known is None:
+                self.solver.add(expr)
+                self._exprs_known[expr] = True
+            elif already_known is not True:
+                raise CrossHairInternal
 
     def rand(self) -> random.Random:
         return self._random
@@ -781,29 +821,37 @@ class StateSpace:
             self._search_position = node
         else:
             node = self._search_position.simplify()
-            assert isinstance(node, ParallelNode)
+            if not isinstance(node, ParallelNode):
+                self.raise_not_deterministic(
+                    node, "Wrong node type (expected ParallelNode)"
+                )
             node._false_probability = false_probability
         self.choices_made.append(node)
         ret, _, next_node = node.choose(self)
         self._search_position = next_node
         return ret
 
-    def is_possible(self, expr: z3.ExprRef) -> bool:
+    def is_possible(self, expr) -> bool:
+        with NoTracing():
+            if hasattr(expr, "var"):
+                expr = expr.var
+            debug("is possible?", expr)
         return solver_is_sat(self.solver, expr)
 
+    def mark_all_parent_frames(self):
+        frames: Set[FrameType] = set()
+        frame = _getframe()
+        while frame and frame not in frames:
+            frames.add(frame)
+            frame = frame.f_back
+        self.external_frames = (
+            frames  # just to prevent dealllocation and keep the id()s valid
+        )
+        self.external_frame_ids = {id(f) for f in frames}
+
     def gen_stack_descriptions(self) -> Tuple[str, ...]:
         f: Any = _getframe().f_back.f_back  # type: ignore
-        f0 = f.f_back or f
-        f1 = f0.f_back or f0
-        f2 = f1.f_back or f1
-        f3 = f2.f_back or f2
-        f4 = f3.f_back or f3
-        f5 = f4.f_back or f4
-        f6 = f5.f_back or f5
-        f7 = f6.f_back or f6
-        frames = (f7, f6, f5, f4, f3, f2, f1, f0)
-        # TODO: if we deprecate 3.7, we could try this instead of the above:
-        # frames = [f := f.f_back or f for _ in range(8)]
+        frames = [f := f.f_back or f for _ in range(8)]
         # TODO: To help oracles, I'd like to add sub-line resolution via f.f_lasti;
         # however, in Python >= 3.11, the instruction pointer can shift between
         # PRECALL and CALL opcodes, triggering our nondeterminism check.
@@ -818,56 +866,44 @@ class StateSpace:
             )
             raise PathTimeout
 
+    @assert_tracing(False)
     def choose_possible(
         self, expr: z3.ExprRef, probability_true: Optional[float] = None
     ) -> bool:
         known_result = self._exprs_known.get(expr)
         if isinstance(known_result, bool):
             return known_result
-        if is_tracing():
-            raise CrosshairInternal
-        self.check_timeout()
-        if self._search_position.is_stem():
-            node = self._search_position.grow_into(
-                WorstResultNode(self._random, expr, self.solver)
-            )
-        else:
-            node = self._search_position.simplify()  # type: ignore
-            assert isinstance(node, WorstResultNode)
-            if not z3.eq(node.expr, expr):
-                debug(" *** Begin Not Deterministic Debug *** ")
-                debug("  Traceback: ", test_stack())
-                debug("Decision expression changed from:")
-                debug(f"  {node.expr}")
-                debug("To:")
-                debug(f"  {expr}")
-                debug(" *** End Not Deterministic Debug *** ")
-                raise NotDeterministic
-
-        self._search_position = node
         # NOTE: format_stack() is more human readable, but it pulls source file contents,
         # so it is (1) slow, and (2) unstable when source code changes while we are checking.
         stacktail = self.gen_stack_descriptions()
-        assert isinstance(node, SearchTreeNode)
-        if not node.stacktail:
+        if self._search_position.is_stem():
+            # We only allow time outs at stems - that's because we don't want
+            # to think about how mutating an existing path branch would work:
+            self.check_timeout()
+            node = self._search_position.grow_into(
+                WorstResultNode(self._random, expr, self.solver)
+            )
+            node.iteration = self._root.iteration
             node.stacktail = stacktail
         else:
-            if node.stacktail != stacktail:
-                debug(" *** Begin Not Deterministic Debug *** ")
-                debug("     First state: ")
-                debug(node.stacktail)
-                debug("     Current state: ")
-                debug(stacktail)
-                debug("     Decision points prior to this:")
-                for choice in self.choices_made:
-                    debug("      ", choice)
-                debug("     Stack Diff: ")
-                import difflib
-
-                debug("\n".join(difflib.context_diff(node.stacktail, stacktail)))
-                debug(" *** End Not Deterministic Debug *** ")
-                raise NotDeterministic
-
+            node = self._search_position.simplify()  # type: ignore
+            not_deterministic_reason = (
+                (
+                    (not isinstance(node, WorstResultNode))
+                    and "Wrong node type (expected WorstResultNode)"
+                )
+                # TODO: Not clear whether we want this stack trace check.
+                # A stack change usually indicates a serious problem, but not 100% of the time.
+                # Keeping it would mean that we fail earlier.
+                # But also see https://github.com/HypothesisWorks/hypothesis/pull/4034#issuecomment-2606415404
+                # or (node.stacktail != stacktail and "Stack trace changed")
+                or ((not z3.eq(node.expr, expr)) and "SMT expression changed")
+            )
+            if not_deterministic_reason:
+                self.raise_not_deterministic(
+                    node, not_deterministic_reason, expr=expr, stacktail=stacktail
+                )
+        self._search_position = node
         choose_true, chosen_probability, stem = node.choose(
             self, probability_true=probability_true
         )
@@ -883,16 +919,48 @@ class StateSpace:
         chosen_expr = expr if choose_true else z3Not(expr)
         if in_debug():
             debug(
-                "SMT chose:",
-                chosen_expr,
-                "(chance:",
-                chosen_probability,
-                ")",
+                f"SMT chose: {chosen_expr} (chance: {chosen_probability}) at",
+                ch_stack(),
             )
-        z3Add(self.solver, chosen_expr)
+        z3Aassert(self.solver, chosen_expr)
         self._exprs_known[expr] = choose_true
         return choose_true
 
+    def raise_not_deterministic(
+        self,
+        node: NodeLike,
+        reason: str,
+        expr: Optional[z3.ExprRef] = None,
+        stacktail: Optional[Tuple[str, ...]] = None,
+        currently_handling: Optional[BaseException] = None,
+    ) -> NoReturn:
+        lines = ["*** Begin Not Deterministic Debug ***"]
+        if hasattr(node, "iteration"):
+            lines.append(f"Previous iteration: {node.iteration}")
+        if hasattr(node, "expr"):
+            lines.append(f"Previous SMT expression: {node.expr}")
+        if expr is not None:
+            lines.append(f"Current SMT expression: {expr}")
+        if not stacktail:
+            if currently_handling is not None:
+                stacktail = tuple(format_tb(currently_handling.__traceback__))
+            else:
+                stacktail = self.gen_stack_descriptions()
+        lines.append("Current stack tail:")
+        lines.extend(f"  {x}" for x in stacktail)
+        if hasattr(node, "stacktail"):
+            lines.append("Previous stack tail:")
+            lines.extend(f"  {x}" for x in node.stacktail)
+        lines.append(f"Reason: {reason}")
+        lines.append("*** End Not Deterministic Debug ***")
+        for line in lines:
+            print(line)
+        exc = NotDeterministic()
+        if currently_handling:
+            raise exc from currently_handling
+        else:
+            raise exc
+
     def find_model_value(self, expr: z3.ExprRef) -> Any:
         with NoTracing():
             while True:
@@ -902,13 +970,13 @@ class StateSpace:
                     )
                 node = self._search_position.simplify()
                 if isinstance(node, SearchLeaf):
-                    raise CrosshairInternal(
+                    raise CrossHairInternal(
                         f"Cannot use symbolics; path is already terminated"
                     )
                 if not isinstance(node, ModelValueNode):
                     debug(" *** Begin Not Deterministic Debug *** ")
                     debug(f"Model value node expected; found {type(node)} instead.")
-                    debug("  Traceback: ", test_stack())
+                    debug("  Traceback: ", ch_stack())
                     debug(" *** End Not Deterministic Debug *** ")
                     raise NotDeterministic
                 (chosen, _, next_node) = node.choose(self, probability_true=1.0)
@@ -924,14 +992,14 @@ class StateSpace:
                     ):
                         self._already_logged.add(expr)
                         debug("SMT realized symbolic:", expr, "==", repr(ret))
-                        debug("Realized at", test_stack())
+                        debug("Realized at", ch_stack())
                     return ret
                 else:
                     self.solver.add(expr != node.condition_value)
 
     def find_model_value_for_function(self, expr: z3.ExprRef) -> object:
         if not solver_is_sat(self.solver):
-            raise CrosshairInternal("model unexpectedly became unsatisfiable")
+            raise CrossHairInternal("model unexpectedly became unsatisfiable")
         # TODO: this need to go into a tree node that returns UNKNOWN or worse
         # (because it just returns one example function; it's not covering the space)
 
@@ -973,7 +1041,7 @@ class StateSpace:
                     continue
                 if self.smt_fork(curref == ref, probability_true=0.1):
                     debug(
-                        "HEAP key lookup ",
+                        "Heap key lookup ",
                         ref,
                         ": Found existing. ",
                         "type:",
@@ -984,7 +1052,7 @@ class StateSpace:
                     return curval
             ret = proxy_generator(typ)
             debug(
-                "HEAP key lookup ",
+                "Heap key lookup ",
                 ref,
                 ": Created new. ",
                 "type:",
@@ -1013,7 +1081,7 @@ class StateSpace:
     def defer_assumption(self, description: str, checker: Callable[[], bool]) -> None:
         self._deferred_assumptions.append((description, checker))
 
-    def detach_path(self) -> None:
+    def detach_path(self, currently_handling: Optional[BaseException] = None) -> None:
         """
         Mark the current path exhausted.
 
@@ -1036,7 +1104,12 @@ class StateSpace:
                 if not prefer_true(check_ret):
                     raise IgnoreAttempt("deferred assumption failed: " + description)
             self.is_detached = True
-            assert self._search_position.is_stem()
+            if not self._search_position.is_stem():
+                self.raise_not_deterministic(
+                    self._search_position,
+                    f"Expect to detach path at a stem node, not at this node: {self._search_position.simplify()}",
+                    currently_handling=currently_handling,
+                )
             node = self._search_position.grow_into(DeatchedPathNode())
             assert node.child.is_stem()
             self.choices_made.append(node)
@@ -1044,6 +1117,8 @@ class StateSpace:
             debug("Detached from search tree")
 
     def cap_result_at_unknown(self):
+        # TODO: this doesn't seem to work as intended.
+        # If any execution path is confirmed, the end result is sometimes confirmed as well.
         self.status_cap = VerificationStatus.UNKNOWN
 
     def bubble_status(
@@ -1064,11 +1139,12 @@ class StateSpace:
             return (analysis, True)
         for node in reversed(self.choices_made):
             node.update_result(analysis)
-        if in_debug():
-            for line in debug_path_tree(
-                self._root, set(self.choices_made + [self._search_position])
-            ):
-                debug(line)
+        if False:  # this is more noise than it's worth (usually)
+            if in_debug():
+                for line in debug_path_tree(
+                    self._root, set(self.choices_made + [self._search_position])
+                ):
+                    debug(line)
         # debug('Path summary:', self.choices_made)
         first = self.choices_made[0]
         analysis = first.get_result()
@@ -1081,3 +1157,4 @@ class StateSpace:
 class SimpleStateSpace(StateSpace):
     def __init__(self):
         super().__init__(monotonic() + 10000.0, 10000.0, RootNode())
+        self.mark_all_parent_frames()
