@@ -1,8 +1,9 @@
 import pathlib
 import sys
+from collections.abc import Container
 from copy import deepcopy
 from dataclasses import dataclass, replace
-from typing import Callable, Iterable, List, Mapping, Optional, Sequence, Set, Tuple
+from typing import Callable, Iterable, List, Mapping, Optional, Sequence, Tuple
 
 from crosshair.core import (
     AnalysisMessage,
@@ -14,16 +15,15 @@ from crosshair.core import (
 )
 from crosshair.options import AnalysisOptionSet
 from crosshair.statespace import context_statespace
-from crosshair.tracers import NoTracing, is_tracing
+from crosshair.tracers import NoTracing, ResumedTracing
 from crosshair.util import (
-    IgnoreAttempt,
-    UnexploredPath,
+    assert_tracing,
+    ch_stack,
     debug,
     in_debug,
+    is_iterable,
     is_pure_python,
     name_of_type,
-    test_stack,
-    true_type,
 )
 
 ComparableLists = Tuple[List, List]
@@ -54,8 +54,8 @@ def check_states(
         )
     elif expected == MessageType.CONFIRMED:
         local_opts = AnalysisOptionSet(
-            per_condition_timeout=10,
-            per_path_timeout=5,
+            per_condition_timeout=60,
+            per_path_timeout=20,
             max_uninteresting_iterations=sys.maxsize,
         )
     elif expected == MessageType.POST_ERR:
@@ -123,9 +123,35 @@ def check_messages(checkables: Iterable[Checkable], **kw) -> ComparableLists:
     return (msgs, [AnalysisMessage(**kw)])
 
 
-def nan_equal(a, b):
+def flexible_equal(a, b):
+    if a is b:
+        return True
+    if type(a) is type(b) and type(a).__eq__ is object.__eq__:
+        # If types match and it uses identity-equals, we can't do much. Assume equal.
+        return True
     if a != a and b != b:  # handle float('nan')
         return True
+    if (
+        is_iterable(a)
+        and not isinstance(a, Container)
+        and is_iterable(b)
+        and not isinstance(b, Container)
+    ):  # unsized iterables compare by contents
+        a, b = list(a), list(b)
+    if type(a) == type(b):
+        # Recursively apply flexible_equal for most containers:
+        if isinstance(a, Mapping):
+            if len(a) != len(b):
+                return False
+            for k, v in a.items():
+                if not flexible_equal(v, b[k]):
+                    return False
+            return True
+        if isinstance(a, Container) and not isinstance(a, (str, bytes)):
+            if len(a) != len(b):
+                return False
+            return all(flexible_equal(ai, bi) for ai, bi in zip(a, b))
+
     return a == b
 
 
@@ -133,6 +159,7 @@ def nan_equal(a, b):
 class ExecutionResult:
     ret: object  # return value
     exc: Optional[BaseException]  # exception raised, if any
+    tb: Optional[str]
     # args after the function terminates:
     post_args: Sequence
     post_kwargs: Mapping[str, object]
@@ -141,7 +168,7 @@ class ExecutionResult:
         if not isinstance(other, ExecutionResult):
             return False
         return (
-            nan_equal(self.ret, other.ret)
+            flexible_equal(self.ret, other.ret)
             and type(self.exc) == type(other.exc)
             and self.post_args == other.post_args
             and self.post_kwargs == other.post_kwargs
@@ -152,7 +179,7 @@ class ExecutionResult:
         if self.exc:
             exc = self.exc
             exc_type = name_of_type(type(exc))
-            tb = test_stack(exc.__traceback__)
+            tb = self.tb or "(missing traceback)"
             ret = f"exc={exc_type}: {str(exc)} {tb}"
         else:
             ret = f"ret={self.ret!r}"
@@ -178,7 +205,8 @@ def summarize_execution(
     if not kwargs:
         kwargs = {}
     ret: object = None
-    exc = None
+    exc: Optional[Exception] = None
+    tbstr: Optional[str] = None
     try:
         possibly_symbolic_ret = fn(*args, **kwargs)
         if detach_path:
@@ -199,13 +227,14 @@ def summarize_execution(
     except Exception as e:
         exc = e
         if detach_path:
-            context_statespace().detach_path()
+            context_statespace().detach_path(e)
         exc = deep_realize(exc)
         # NOTE: deep_realize somehow empties the __traceback__ member; re-assign it:
         exc.__traceback__ = e.__traceback__
+        tbstr = ch_stack(currently_handling=exc)
         if in_debug():
-            debug("hit exception:", type(exc), exc, test_stack(exc.__traceback__))
-    return ExecutionResult(ret, exc, args, kwargs)
+            debug("hit exception:", type(exc), exc, tbstr)
+    return ExecutionResult(ret, exc, tbstr, args, kwargs)
 
 
 @dataclass
@@ -235,8 +264,8 @@ def compare_returns(fn: Callable, *a: object, **kw: object) -> ResultComparison:
     return comparison
 
 
+@assert_tracing(True)
 def compare_results(fn: Callable, *a: object, **kw: object) -> ResultComparison:
-    assert is_tracing()
     original_a = deepcopy(a)
     original_kw = deepcopy(kw)
     symbolic_result = summarize_execution(fn, a, kw)
@@ -245,16 +274,25 @@ def compare_results(fn: Callable, *a: object, **kw: object) -> ResultComparison:
     concrete_kw = deep_realize(original_kw)
 
     # Check that realization worked, too:
-    for idx, arg in enumerate(concrete_a):
-        if true_type(arg) != type(arg):
+    with NoTracing():
+        labels_and_args = [
+            *(
+                (f"Argument {idx + 1}", a[idx], arg)
+                for idx, arg in enumerate(concrete_a)
+            ),
+            *((f"Keyword argument '{k}'", kw[k], v) for k, v in concrete_kw.items()),
+        ]
+        for label, symbolic_arg, concrete_arg in labels_and_args:
+            with ResumedTracing():
+                symbolic_type = type(symbolic_arg)
+                concrete_type = type(concrete_arg)
+            true_concrete_type = type(concrete_arg)
             assert (
-                False
-            ), f"Argument {idx + 1} was {true_type(arg)} afer realization; expected {type(arg)}"
-    for k, v in concrete_kw.items():
-        if true_type(v) != type(v):
+                true_concrete_type == concrete_type
+            ), f"{label} did not realize. It is {true_concrete_type} instead of {concrete_type}."
             assert (
-                False
-            ), f"Keyword argument '{k}' was {true_type(v)} afer realization; expected {type(arg)}"
+                true_concrete_type == symbolic_type
+            ), f"{label} should realize to {symbolic_type}; it is {true_concrete_type} instead."
 
     with NoTracing():
         concrete_result = summarize_execution(
